import React, { useEffect, useState } from "react";
import { SessionInfoPlayResultOver, useSession } from "../lib/store"
import { CPUNAME, CPUS_NAMES } from "../lib/CPUS";
import { Play, PLAYER_OPTIONS, PlayerAction, PlayResult } from "../lib/game";
import './Game.css';

// Future improvements create a GameBoard that controls the size of the boxes 
// rather than having the code here multiple times
export function Game() {
    const session = useSession();

    const [selectedCPU, setSelectedCPU] = useState(CPUS_NAMES[0]);

    const [gameState, setGameState] = useState<'play' | 'result' | 'over'>('play');
    const [result, setResult] = useState<Play | undefined>();
    // TODO move this to a type
    const [lastGame, setLastGame] = useState<SessionInfoPlayResultOver | undefined>();

    // If reset or logout is hit reset the state of this component
    useEffect(() => {
        if (session.currentUser === undefined) {
            setLastGame(undefined);
            setResult(undefined);
            setGameState('play');
        }
    }, [session.currentUser]);

    if (!session.currentUser) {
        return <div className="w-full flex flex-col justify-center items-center p-5">
            <span className="text-4xl">↑</span>
            <h2 className="text-2xl font-bold">Please choose a name before starting</h2>
        </div>
    }

    if (gameState === 'over') {
        if (!lastGame) {
            // This should not happen
            setGameState('play');
            return <></>
        }

        // TODO maybe add an icon on top of the winner
        return <>
            <GameBoard
                shrinkLoser={lastGame.winner}
                playerSide={<span className="text-4xl whitespace-nowrap">{session.currentUser}</span>}
                cpuSide={<span className="text-4xl whitespace-nowrap">{lastGame.game.cpu}</span>}
            >
                {lastGame.winner === 'tie' && <span className="top-1/2 left-1/2 -translate-1/2 text-yellow-400 drop-shadow-lg drop-shadow-yellow-400 absolute text-6xl">It&#39;s a Tie</span>}
                <button
                    className="defaultBtn absolute bottom-[2em] left-1/2 -translate-x-1/2 shadow-md"
                    onClick={() => {
                        setGameState('play')
                    }}
                >
                    New Game
                </button>
            </GameBoard>
            <GameHistory plays={lastGame.game.plays} />
        </>
    }

    if (!session.currentGame) {
        return <GameBoard
            playerSide={<span className="text-3xl">{session.currentUser}</span>}
            cpuSide={
                <select className="text-3xl" value={selectedCPU} onChange={(e) => setSelectedCPU(e.currentTarget.value as CPUNAME)}>
                    {CPUS_NAMES.map(a => <option key={a} value={a}>{a}</option>)}
                </select>
            }
        >
            <button
                className="defaultBtn absolute bottom-[2em] left-1/2 -translate-x-1/2 shadow-md"
                onClick={() => session.startGame(selectedCPU)}
            >
                Start Game
            </button>
            <span className="absolute top-1/2 left-1/2 -translate-x-[56%] -translate-y-1/2 drop-shadow-lg font-bold text-6xl">VS</span>
        </GameBoard>
    }

    if (gameState === 'result') {
        if (!result) {
            // This should not happen
            setGameState('play');
            return <></>
        }
        return <>
            <GameBoard
                playerSide={<>
                    <span className="absolute top-[1em] left-1/2 -translate-1/2 text-2xl">{session.currentUser}</span>
                    <div className="text-2xl">
                        <span className="font-bold">
                            {result.user}
                        </span>
                        <WinnerText me='user' result={result.winner} />
                    </div>
                </>}
                cpuSide={<>
                    <span className="absolute bottom-2 md:bottom-[unset] md:top-[1em] left-1/2 -translate-x-1/2 text-2xl whitespace-nowrap">{session.currentGame!.cpu}</span>
                    <div className="text-2xl">
                        <span className="font-bold">
                            {result.cpu}
                        </span>
                        <WinnerText me='cpu' result={result.winner} />
                    </div>
                </>}
            >
                <span className="absolute top-1/2 left-1/2 -translate-x-[56%] -translate-y-1/2 drop-shadow-lg font-bold text-6xl">VS</span>
                <button
                    className="defaultBtn absolute text-xs px-2! md:text-base right-1 top-1/2 md:top-[unset] -translate-x-0 md:right-[unset] md:-translate-x-1/2 md:bottom-[2em] md:left-1/2 -translate-1/2 shadow-md"
                    onClick={() => {
                        setGameState('play')
                    }}
                >
                    Next Round
                </button>
                <PointsDisplay />
            </GameBoard>
            <GameHistory />
        </>
    }

    function play(action: PlayerAction) {
        const r = session.play(action);
        if (r.type === 'play') {
            setGameState('result');
            setResult(r.play);
        } else {
            setGameState('over');
            setLastGame(r);
        }
    }

    if (gameState === 'play') {
        return <>
            <GameBoard
                playerSide={
                    <>
                        <span className="absolute top-[1em] left-1/2 -translate-1/2 text-2xl">{session.currentUser}</span>
                        <div className="flex flex-wrap justify-around gap-2">
                            {PLAYER_OPTIONS.map((a) =>
                                <button
                                    key={a}
                                    type="button"
                                    className="p-2 drop-shadow-lg text-md md:text-2xl hover:drop-shadow-white cursor-pointer"
                                    onClick={() => play(a)}
                                >
                                    {a}
                                </button>
                            )}
                        </div>
                    </>
                }
                cpuSide={<span className="text-4xl">{session.currentGame.cpu}</span>}
            >
                <span className="absolute top-1/2 left-1/2 -translate-x-[56%] -translate-y-1/2 drop-shadow-lg font-bold text-6xl">VS</span>
                <PointsDisplay />
            </GameBoard>
            <GameHistory />
        </>
    }

    return <></>;
}

function PointsDisplay() {
    const session = useSession();
    const icons = Array(3).fill(null);
    return <div className="absolute flex gap-2 bg-white/20 rounded-lg shadow-lg p-2 left-2 top-1/2 -translate-y-1/2 md:top-2 md:left-1/2 md:-translate-x-1/2 md:translate-y-0">
        {icons.map((_, i) => <span key={i} className={
            `${session.currentGame?.plays[i] === undefined ? 'bg-gray-500' :
                session.currentGame?.plays[i].winner === 'cpu' ? 'bg-red-500' :
                    session.currentGame?.plays[i].winner === 'user' ? 'bg-blue-500' :
                        'bg-yellow-500'
            } border-solid border border-black rounded-full min-h-[15px] min-w-[15px]`
        } />)}
    </div>
}

function WinnerText({ me, result }: { me: PlayResult, result: PlayResult }) {
    let r = <span>Loser</span>;
    if (result === 'tie') r = <span className="text-yellow-500">Tie</span>;
    else if (result === me) r = <span>Winner</span>;
    return <div className="text-2xl">
        {r}
    </div>
}

function GameHistory({ plays }: { plays?: Play[] }) {
    const session = useSession();

    if ((!session.currentGame || session.currentGame.plays.length === 0) && plays === undefined) {
        return <></>
    }

    return <>
        <h2 className="font-bold pt-5">
            Game History
        </h2>
        <div>
            {(plays ?? session.currentGame!.plays).map((play, i) => <div key={i} className="p-2">
                User played <span className="text-blue-500">{play.user}</span>, and CPU played <span className="text-red-500">{play.cpu}</span>! Winner:{" "}
                <span className={
                    play.winner === 'tie' ? 'text-yellow-500' :
                        play.winner === 'cpu' ? 'text-red-500' :
                            'text-blue-500'
                }>{play.winner}</span>
            </div>)}
        </div>
    </>
}

function GameBoard({
    cpuSide,
    playerSide,
    children,
    shrinkLoser
}: {
    cpuSide?: React.ReactNode,
    playerSide?: React.ReactNode,
    children?: React.ReactNode,
    shrinkLoser?: PlayResult
}) {
    function getBoxClass(target: PlayResult) {
        if (!shrinkLoser) return 'h-1/2 w-full md:w-1/2 md:h-full flex-grow'
        if (shrinkLoser === 'tie') return 'h-full w-full';
        if (shrinkLoser !== target) return 'loser-shrink-to-zero h-0 md:h-full w-full md:w-0';
        return 'h-1/2 w-full md:w-1/2 md:h-full flex-grow'
    }

    return <div className="h-[500px] flex flex-col md:flex-row rounded-xl overflow-hidden relative">
        <div className={`bg-blue-500/50 grid place-items-center overflow-hidden relative ${getBoxClass('user')}`}>
            {playerSide}
        </div>
        <div className={`bg-red-500/50 grid place-items-center overflow-hidden relative ${getBoxClass('cpu')}`}>
            {cpuSide}
        </div>
        {children}
    </div >
}
